#include "lib.h"

TypeTree DB_TYPE;
NameTree DB_NAME;
Pool<Variable> DB_DATA;

extern "C" uint32_t type_outer(uint32_t id, uint32_t key)
{
    return DB_TYPE.outer(id, key);
}

extern "C" uint32_t type_inner(uint32_t id)
{
    return DB_TYPE.inner(id);
}

extern "C" uint32_t type_key(uint32_t id)
{
    return DB_TYPE.value(id);
}

extern "C" size_t type_size(uint32_t type_id)
{
    if(DB_TYPE.has(type_id)) {
        uint32_t type = DB_TYPE.value(type_id);

        switch(type) {
            case Type::Tag::Varying: return sizeof(Variable);
            case Type::Tag::Null:    return 0;
            case Type::Tag::Boolean: return sizeof(Type::Boolean);
            case Type::Tag::Natural: return sizeof(Type::Natural);
            case Type::Tag::Integer: return sizeof(Type::Integer);
            case Type::Tag::Block:   return DB_TYPE.value(DB_TYPE.inner(type_id));
            case Type::Tag::String:  return sizeof(Type::List);
            case Type::Tag::Array: {
                uint32_t inner = static_cast<uint32_t>(DB_TYPE.inner(type_id));
                return static_cast<size_t>(DB_TYPE.value(inner)) * type_size(static_cast<uint32_t>(DB_TYPE.inner(inner)));
            };
            case Type::Tag::List:    return sizeof(Type::List);
            default: return 0;
        }
    }
    return 0;
}

void* allocate(uint32_t type_id, size_t count)
{
    void* mem = nullptr;
    size_t size = type_size(type_id) * count;
    if(size > 0) {
        mem = malloc(size);
        if(mem != nullptr) {
            memset(mem, 0, size);
        }
        return mem;
    }
    return nullptr;
}

extern "C" size_t acquire(uint32_t type_id)
{
    void* mem = allocate(type_id, 1);
    Variable addr {0};
    if(mem != nullptr) {
        addr.type = type_id;
        addr.address = mem;
        return DB_DATA.add(addr);
    }
    return 0;
 }

extern "C" bool release(size_t id)
{
    if(DB_DATA.has(id)) {
        Variable addr;
        if(DB_DATA.get(id, addr)) {
            if(addr.address != nullptr) {
                free(addr.address);
            }
            DB_DATA.remove(id);
            return true;
        }
    }
    return false;
}

extern "C" Variable address(size_t id)
{
    Variable addr {0};
    if(DB_DATA.has(id)) {
        DB_DATA.get(id, addr);
    }
    return addr;
}

bool copy(Variable src, Variable dst)
{
    Variable source = src;
    Variable destination = dst;
    
    // dereference varying data
    if(type_key(src.type) == Type::Tag::Varying) {
        source = *reinterpret_cast<Variable*>(src.address);
    }

    // prepare destination for varying data
    if(type_key(dst.type) == Type::Tag::Varying) {
        auto& dest_ref = *reinterpret_cast<Variable*>(dst.address);
        
        // determine if memory can be reused, otherwise free and reallocate
        if(source.type != dest_ref.type) {
            if(dest_ref.address != nullptr) {
                free(dest_ref.address);
                dest_ref.type = Type::Tag::Null;
                dest_ref.address = nullptr;
            }
            dest_ref.type = source.type;
            dest_ref.address = allocate(dest_ref.type, 1);
        }
    }

    // copy data into destination
    if(source.type == destination.type) {
        switch(type_key(destination.type)) {
            case Type::Tag::Null: { } break;

            case Type::Tag::Boolean:
            case Type::Tag::Natural:
            case Type::Tag::Integer:
            case Type::Tag::Block:
            {
                memcpy(destination.address, source.address, type_size(source.type));
            } break;
            
            case Type::Tag::List: {
                auto& src_list = *reinterpret_cast<Type::List*>(source.address);
                auto& dst_list = *reinterpret_cast<Type::List*>(destination.address);
                
                dst_list.capacity = src_list.capacity;
                dst_list.length = src_list.length;
                dst_list.data = allocate(type_inner(source.type), src_list.capacity);

                for(size_t i = 0; i < src_list.length; ++i) {
                    copy(list_at(source, i), list_at(destination, i));
                }
            } break;
        }

        return true;
    }

    return false;
}


// Boolean //

extern "C" void bool_set(Variable addr, Type::Boolean value)
{
    if(type_key(addr.type) == Type::Tag::Boolean) {
        *(reinterpret_cast<Type::Boolean*>(addr.address)) = value;
    }
}

extern "C" Type::Boolean bool_get(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::Boolean) {
        return *(reinterpret_cast<Type::Boolean*>(addr.address));
    }
    return false;
}


// Natural //

extern "C" void natural_set(Variable addr, Type::Natural value)
{
    if(type_key(addr.type) == Type::Tag::Natural) {
        *(reinterpret_cast<Type::Natural*>(addr.address)) = value;
    }
}

extern "C" Type::Natural natural_get(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::Natural) {
        return *(reinterpret_cast<Type::Natural*>(addr.address));
    }
    return false;
}


// Integer //

extern "C" void integer_set(Variable addr, Type::Integer value)
{
    if(type_key(addr.type) == Type::Tag::Integer) {
        *(reinterpret_cast<Type::Integer*>(addr.address)) = value;
    }
}

extern "C" Type::Integer integer_get(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::Integer) {
        return *(reinterpret_cast<Type::Integer*>(addr.address));
    }
    return false;
}


// List

extern "C" size_t list_capacity(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::List) {
        return (*reinterpret_cast<Type::List*>(addr.address)).capacity;
    }
    return 0;
}

extern "C" size_t list_length(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::List) {
        return (*reinterpret_cast<Type::List*>(addr.address)).length;
    }
    return 0;
}

extern "C" Variable list_first(Variable addr)
{
    Variable result {0};
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        if(list.length > 0) {

        }
    }
    return result;
}

extern "C" Variable list_last(Variable addr)
{
    Variable result {0};
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        if(list.length > 0) {

        }
    }
    return result;
}

void* list_cell(Variable addr, size_t index)
{
    if(type_key(addr.type) == Type::Tag::List) {
        uint32_t inner = type_inner(addr.type);
        Type::List& list = *reinterpret_cast<Type::List*>(addr.address);

        // validate for overflow
        size_t offset = type_size(inner);
        if(list.data != nullptr && offset > 0 && index < list.length) {
            return reinterpret_cast<void*>(reinterpret_cast<size_t>(addr.address) + (offset * index));
        }
    }
    return nullptr;
}

extern "C" Variable list_at(Variable addr, size_t index)
{
    Variable result {0};
    if(type_key(addr.type) == Type::Tag::List) {
        uint32_t inner = type_inner(addr.type);
        Type::List& list = *reinterpret_cast<Type::List*>(addr.address);

        void* cell = list_cell(addr, index);
        if(cell != nullptr) {
            // handle varying inner type
            if(type_key(inner) == Type::Tag::Varying) {
                result = *reinterpret_cast<Variable*>(cell);
            } else {
                result.type = inner;
                result.address = cell;
            }
        }
    }
    return result;
}

extern "C" void list_clear(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        list.length = 0;
        memset(list.data, 0, list.offset * list.capacity);
    }
}

extern "C" void list_prepend(Variable addr, Variable source)
{
    if(type_key(addr.type) == Type::Tag::List) {
        uint32_t inner = type_inner(addr.type);

        // validate list can store value
        if(type_key(inner) == Type::Tag::Varying || source.type == inner) {
            auto& list = (*reinterpret_cast<Type::List*>(addr.address));

            if(list.length == list.capacity) {
                size_t new_capacity = (list.capacity < 16)? 16 : list.capacity * 2;
                void* new_data = malloc(list.offset * new_capacity);
                memset(new_data, 0, list.offset * new_capacity);
                memcpy(new_data, list.data, list.offset * list.capacity);
            }

            // copy source to cell
            Variable dst {0};
            dst.type = inner;
            dst.address = reinterpret_cast<void*>(reinterpret_cast<size_t>(addr.address) + (list.offset * list.length));
            copy(source, dst);

            list.length++;
        }
    }
}

extern "C" void list_append(Variable addr, Variable source)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}

extern "C" void list_insert(Variable addr, size_t index, Variable source)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}

extern "C" void list_truncate(Variable addr, size_t maximum)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}

extern "C" void list_shift(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}

extern "C" void list_remove(Variable addr, size_t index, size_t maximum)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}

extern "C" void list_reserve(Variable addr, size_t capacity)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));

        if(list.capacity < capacity) {
            void* new_data = allocate(type_inner(addr.type), capacity); //malloc(list.offset * capacity);
            //memset(new_data, 0, list.offset * capacity);
            memcpy(new_data, list.data, list.offset * list.capacity);
        }
    }
}

extern "C" void list_resize(Variable addr, size_t length)
{
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}

